#include <stdlib.h>
#include <stdio.h>

#include <string.h>

#ifdef _WIN32
#include <windows.h>
#endif

#include <openssl/bn.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/ui.h>

#include "qrc.h"

#include "qr.h"


#define ECC 0
#define VERSION 10


static unsigned char *g_binbuf = NULL;
static unsigned long g_binsize = 0, g_binpos = 0;


void die(const char *msg) {

  fprintf(stderr, "%s\n", msg);

  exit(-1);
}


bool QRC_go(CQR_Encode *qrc) {
  unsigned char *p;
  unsigned long size;
  bool r = false;

  size = (g_binsize - g_binpos);
  p = new unsigned char[4 + size];
  for(int i = 0; i < 4; i++) p[i] = (unsigned char)(g_binsize >> (24 - (8 * i)) & 0xFF);
  memcpy((p + 4), (g_binbuf + g_binpos), size);

  for(int i = size; i > 0; i--) {
    r = (qrc->EncodeData(ECC, VERSION, FALSE, -1, (BYTE*)p, (4 + i)) != FALSE);
    if(r) {
      size = i;
      break;
    }
  }

  g_binpos += size;

  free(p);

  return(r);
}


static void QRC_bmp(const char *basename) {
  CQR_Encode qrc;

  for(int fi = 0; g_binpos < g_binsize; fi++) {
    char filename[1000];
    int ss, bs;
    FILE *fp;
    if(!QRC_go(&qrc)) die("QR-code generation failed.");

    sprintf(filename, "%s-%d.bmp", basename, fi);
    fp = fopen(filename, "wb");
    if(fp == NULL) die("cannot open output file.");

    ss = (4 + (4 * qrc.m_nSymbleSize) + 4);
    bs = (4 * ss * ss);

    {
      BITMAPFILEHEADER bmfh;
      memset(&bmfh, 0, sizeof(bmfh));
      bmfh.bfType = 'MB';
      bmfh.bfSize = (sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + bs);
      bmfh.bfOffBits = (sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER));
      fwrite(&bmfh, sizeof(bmfh), 1, fp);
    }

    {
      BITMAPINFOHEADER bmih;
      memset(&bmih, 0, sizeof(bmih));
      bmih.biSize = sizeof(bmih);
      bmih.biWidth = ss;
      bmih.biHeight = ss;
      bmih.biPlanes = 1;
      bmih.biBitCount = 32;
      bmih.biCompression = BI_RGB;
      bmih.biSizeImage = bs;
      bmih.biClrUsed =0;
      bmih.biClrImportant = 0;
      fwrite(&bmih, sizeof(bmih), 1, fp);
    }

    {
      DWORD *t;
      t = new DWORD[ss];
      for(int i = 0; i < ss; i++) t[i] = 0xFFFFFF;
      for(int y = 0; y < 4; y++) fwrite(t, 4, ss, fp);
      for(int y = 0; y < qrc.m_nSymbleSize; y++) {
        int k = 0;
        for(int i = 0; i < 4; i++) t[k++] = 0xFFFFFF;
        for(int x = 0; x < qrc.m_nSymbleSize; x++) {
          if(qrc.m_byModuleData[y][x]) {
            for(int i = 0; i < 4; i++) t[k++] = 0x000000;
          } else {
            for(int i = 0; i < 4; i++) t[k++] = 0xFFFFFF;
          }
        }
        for(int i = 0; i < 4; i++) t[k++] = 0xFFFFFF;
        for(int i = 0; i < 4; i++) fwrite(t, 4, ss, fp);
      }
      for(int i = 0; i < ss; i++) t[i] = 0xFFFFFF;
      for(int y = 0; y < 4; y++) fwrite(t, 4, ss, fp);
      delete(t);
    }

    fclose(fp);
  }

  return;
}


void KEY_add(const void *p, unsigned long s) {

  memcpy((g_binbuf + g_binsize), p, s);
  g_binsize += s;

  return;
}


void KEY_u32(unsigned long x) {
  unsigned char t[4];

  // NOTE: JavaVM is big endian.
  for(int i = 0; i < 4; i++) t[i] = (unsigned char)(x >> (24 - (8 * i)));

  KEY_add(t, 4);

  return;
}


void KEY_string(const char *x) {
  int s;

  s = strlen(x);

  KEY_u32(s);
  KEY_add(x, s);

  return;
}


void KEY_bn(BIGNUM *bignum) {
  unsigned char t[65536];
  int s, nh;

  s = (BN_num_bytes(bignum) + 1);

  t[0] = 0;
  BN_bn2bin(bignum, (t + 1));

  nh = ((t[1] & 0x80) ? 0 : 1);
  s -= nh;

  KEY_u32(s);
  KEY_add((t + nh), s);

  return;
}


static int pwdcb(char *p, int size, int verify, void *keyname) {

  if(verify) return(strlen(p));

  {
    UI *ui;
    char prompt[1024];
    int ok;
    ui = UI_new_method(UI_OpenSSL());
    sprintf(prompt, "Enter passphrase for key '%s': ", (const char*)keyname);
    UI_add_input_string(ui, prompt, UI_INPUT_FLAG_DEFAULT_PWD, p, 1, (size - 1));
    do {
      ok = UI_process(ui);
    } while((ok < 0) && UI_ctrl(ui, UI_CTRL_IS_REDOABLE, 0, 0, 0));
    if(ok < 0) {
      size = -1;
    } else {
      size = strlen(p);
    }
  }

  return(size);
}

#ifndef MAX_PATH
#define MAX_PATH 1024
#endif

int main(int argc, char *argv[]) {
  const char *key_filename = NULL;
  char out_basename[MAX_PATH];
  char outmode = 'q';
  RSA *rsa;

  g_binbuf = new unsigned char[65536];

  out_basename[0] = 0;
  out_basename[MAX_PATH - 1] = 0;

  {
    bool options = true;
    for(int i = 1; i < argc; i++) {
      if(options && (argv[i][0] == '-')) {
        switch(argv[i][1]) {
        case '-':
          options = false;
          break;
        case 0:
          key_filename = "-";
          strcpy(out_basename, "id");
          break;
        case 'o':
          strncpy(out_basename, argv[++i], MAX_PATH);
          if(out_basename[MAX_PATH - 1] != 0) die("output basename is too long!");
          break;
        case 'b':
          outmode = 'b';
          break;
        }
      } else {
        if(key_filename == NULL) {
          key_filename = argv[i];
        }
      }
    }
  }

  if(key_filename == NULL) {
    fprintf(stdout, "usage: %s [OPTIONS] (private key filename)\n", argv[0]);
    fprintf(stdout, "\n");
    fprintf(stdout, "[OPTIONS]\n");
    fprintf(stdout, "\t-o (output file name):\n");
    fprintf(stdout, "\t\tspecify a output basename. default filename is 'id'.\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "\t-b (binary output):\n");
    fprintf(stdout, "\t\tspecify to output a raw binary data. in default, output is QR-code bitmap.\n");
    exit(-1);
  }

  if(out_basename[0] == 0) {
    strncpy(out_basename, key_filename, MAX_PATH);
    if(out_basename[MAX_PATH - 1] != 0) die("please specify output basename. input key name is too long!");
    for(int i = strlen(out_basename); i >= 0; i--) {
      if(out_basename[i] == '.') {
        out_basename[i] = 0;
        break;
      }
      if(out_basename[i] == '\\') {
        break;
      }
      if(out_basename[i] == '/') {
        break;
      }
    }
  }

#if 0
#ifdef _WIN32
  {
    char t[MAX_PATH + 2];
    int i;
    i = GetModuleFileName(NULL, t, MAX_PATH);
    while(--i > 0) {
      if(t[i] == '\\') break;
    }
    t[i] = 0;
    SetCurrentDirectory(t);
  }
#endif
#endif

  OpenSSL_add_all_algorithms();

  {
    BIO *bio;
    EVP_PKEY *pk;
    bio = BIO_new(BIO_s_file());
    if(strcmp(key_filename, "-") != 0) {
      if(!BIO_read_filename(bio, key_filename)) die("cannot open private key file");
    } else {
      BIO_set_fp(bio, stdin, BIO_NOCLOSE);
    }
    pk = PEM_read_bio_PrivateKey(bio, NULL, pwdcb, (void*)key_filename);
    if(pk == NULL) die("cannot load private key.");
    rsa = EVP_PKEY_get1_RSA(pk);
    EVP_PKEY_free(pk);
    BIO_free(bio);
  }

  KEY_string("ssh-rsa");
  if(rsa->e == NULL) die("e is not found!");
  KEY_bn(rsa->e);
  if(rsa->n == NULL) die("n is not found!");
  KEY_bn(rsa->n);
  if(rsa->d == NULL) die("d is not found!");
  KEY_bn(rsa->d);

  switch(outmode) {
  case 'b':
    {
      char filename[1024];
      FILE *fp;
      sprintf(filename, "%s.bin", out_basename);
      fp = fopen(filename, "wb");
      if(fp == NULL) die("cannot open output file.");
      fwrite(g_binbuf, g_binsize, 1, fp);
      fclose(fp);
    }
    break;
  default:
    QRC_bmp(out_basename);
    break;
  }

  RSA_free(rsa);

  return(0);
}
